#include <stdint.h>
#include <alloca.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <X11/Xatom.h>
#include "gfx.h"


static Display *x_display;
static Window   x_root_window;
static int      x_screen_number;
static uint32_t x_root_width, x_root_height;
static uint32_t x_red_mask, x_red_shift,
                x_green_mask, x_green_shift,
                x_blue_mask, x_blue_shift;
static int      x_depth;
static Visual  *x_visual;

static int      x_window;
static Pixmap   x_pixmap;
static int      x_window_width, x_window_height;

static int      exposock=-1;
static pid_t    expopid;
static  Atom    wm_protocols, wm_delete_window;

static int fork_exposure_process();
static int send_child_params(Window window, Pixmap pixmap,
                             int width, int height);
static void init_button_colors();

/***********************************************************************
  Images can be shared on non-shared. They are represented by the
  following struct.
 ***********************************************************************/
typedef struct 
{
  uint32_t *data;
  unsigned int width,height, bpl;
  int shared;
  XShmSegmentInfo shminfo;
  XImage *x11image;
} ximage_t;

static int       can_use_shared_images();
static ximage_t *ximage_create_simple(uint32_t width,uint32_t height);
static ximage_t *create_image(uint32_t width, uint32_t height);
static void      destroy_image(ximage_t *image);
static void      put_image(Drawable draw, int dest_x, int dest_y,
                           ximage_t *image, int src_x, int src_y,
                           int width, int height);

static void checker_pattern(ximage_t *img,int N);
static int gfx_recsel_event(XEvent *event);
static int button_event(XEvent *event, gfx_event_t *R);

/***********************************************************************
  We have only one image.
 ***********************************************************************/
static ximage_t *x_image;
static uint32_t **x_image_data;

/***********************************************************************
  In order to create images etc, we need to find a suitable visual 
  with depth==24 or depth==32.
 ***********************************************************************/

static uint32_t trailing_zero_bits(uint32_t Y)
{
  uint32_t shif;
  for(shif=0;Y%2==0;shif++) Y= Y>>1;
  return shif;
}

static int get_x_visual()
{
  int depth, ndepths, i;
  int *depths;
  XVisualInfo xvinfo;

  depths= XListDepths(x_display, x_screen_number, &ndepths);
  if (!depths) 
  {
    fprintf(stderr,"can't get depth list for display\n");
    return 1;
  }

  depth=0;
  for(i=0;i<ndepths;i++)
    if (depths[i]==32 || depths[i]==24)
    {
      depth= depths[i];
      break;
    }
  XFree(depths);

  if (!depth)
  {
    fprintf(stderr, "display doesn't support 24 or 32 bit depth.\n");
    return 1;
  }

  x_depth= depth;
  x_visual= NULL;
  if (XMatchVisualInfo(x_display, x_screen_number, depth, TrueColor,&xvinfo))
     x_visual= xvinfo.visual;
  else if (XMatchVisualInfo(x_display, x_screen_number,
                                       depth, DirectColor,&xvinfo))
     x_visual= xvinfo.visual;
  else
  {
    fprintf(stderr,"can't find a 24 or 32 bit TrueColor/DirectColor visual\n");
    return 1;
  }
  x_red_mask= x_visual->red_mask;
  x_red_shift= trailing_zero_bits(x_red_mask);
  x_green_mask= x_visual->green_mask;
  x_green_shift= trailing_zero_bits(x_green_mask);
  x_blue_mask= x_visual->blue_mask;
  x_blue_shift= trailing_zero_bits(x_blue_mask);
  return 0;
}

static Window create_window(char *title,int width, int height)
{
  int                  x,y;
  Window               window;
  XSetWindowAttributes attribs;
  XTextProperty        title_prop;
  XSizeHints          *size;

  x= (x_root_width-width)/2;
  y= (x_root_height-height)/2;

  attribs.background_pixmap= None;

  window= XCreateWindow(x_display, x_root_window,
                        x, y, width, height,
                        1, x_depth, InputOutput, x_visual, 
                        CWBackPixmap, &attribs);

  if (XStringListToTextProperty(&title, 1, &title_prop))
  {
    XSetWMName(x_display, window, &title_prop);
  }
  size= XAllocSizeHints();
  size->flags= PMinSize | PMaxSize;
  size->min_width= size->max_width= width;
  size->min_height= size->max_height= height;
  XSetWMNormalHints(x_display,window, size);

  unsigned long protos[4];

  wm_protocols= XInternAtom(x_display, "WM_PROTOCOLS", False);
  wm_delete_window= XInternAtom(x_display, "WM_DELETE_WINDOW", False);
  protos[0]= wm_delete_window;
  protos[1]= 0;
 
  XChangeProperty(x_display, window, wm_protocols, XA_ATOM, 32,
                  PropModeReplace, (unsigned char*)protos, 1);

  XFlush(x_display);
  XFree(size);

  return window;
}

static void get_root_attribs()
{
  x_screen_number= DefaultScreen(x_display);
  x_root_window= DefaultRootWindow(x_display);
  x_root_width= DisplayWidth(x_display, x_screen_number);
  x_root_height= DisplayHeight(x_display, x_screen_number);
}


int gfx_init(char *name, int width, int height, char *title)
{
  int y;
  fork_exposure_process();
  x_display= XOpenDisplay(name);

  if (!x_display) 
  {
    //FIXME: log this
    fprintf(stderr,"can't open display\n");
    return 1;
  }
  get_root_attribs();
  if (get_x_visual())
  {
    XCloseDisplay(x_display);
    x_display= NULL;
    return 1;
  }
  x_window= create_window(title,width, height);
  XSelectInput(x_display, x_window, PointerMotionMask | KeyPressMask |
                                    KeyReleaseMask | ButtonPressMask |
                                    ButtonReleaseMask);

  x_pixmap= XCreatePixmap(x_display, x_window, width, height, x_depth);
  x_image= create_image(width, height);

  x_image_data= malloc(sizeof(uint32_t*) * height);
  for(y=0;y<height;y++)
    x_image_data[y]= (uint32_t*) ((char*) x_image->data + y*x_image->bpl);
  checker_pattern(x_image, 17);
  send_child_params(x_window,x_pixmap, width, height);
  x_window_width= width;
  x_window_height= height;
  init_button_colors();
  gfx_update_all();
  XMapWindow(x_display, x_window);
  return 0;
}


/***********************************************************************
 *                                                                     *
 *                           Image Functions                           *
 *                                                                     *
 ***********************************************************************/

static ximage_t *create_image(uint32_t width, uint32_t height)
{
  ximage_t *image;
  size_t data_size;
  struct shmid_ds dummy_shmid_ds;

  if (!can_use_shared_images()) goto fail_nolib;

  image= calloc(1,sizeof(*image));
  image->width= width;
  image->height= height;
  image->shared= 1;

  // Create the image on the server
  //
  image->x11image= XShmCreateImage(x_display,x_visual, x_depth, ZPixmap,
                                   NULL, &image->shminfo, width, height);
  if (!image->x11image) goto fail_create;

  image->bpl= image->x11image->bytes_per_line;
  data_size= height*image->bpl;

  // Create the shared memory area
  //
  image->shminfo.shmid= shmget(IPC_PRIVATE, data_size, IPC_CREAT | 0777);
  if (image->shminfo.shmid <0)
  {
    fprintf(stderr,"can't shmget shared memory\n");
    goto fail_shmcreate;
  }
  
  image->shminfo.shmaddr= shmat(image->shminfo.shmid,0,0);
   // we remove the ID now, so that when the program exits, the memory
   // will be automatically freed
   //
  shmctl(image->shminfo.shmid, IPC_RMID, &dummy_shmid_ds); 
  if (image->shminfo.shmaddr==(char*) -1) 
  {
    fprintf(stderr,"can't attach shared memory\n");
    goto fail_shmat;
  }

  // bind the area to the image 
  //
  image->data= (uint32_t*) image->shminfo.shmaddr;
  image->x11image->data= (char*) image->shminfo.shmaddr;
  if (!XShmAttach(x_display, &image->shminfo)) 
  {
    fprintf(stderr,"couldn't attach shared memory to the ximage\n");
    goto fail_attach_x;
  }

  return image;

fail_attach_x:
  shmdt(image->shminfo.shmaddr);
fail_shmat:      // We have already removed the ID
fail_shmcreate:
  XDestroyImage(image->x11image);
fail_create:
  free(image);
fail_nolib:
  return ximage_create_simple(width,height);
}



static ximage_t *ximage_create_simple(uint32_t width,uint32_t height)
{
  ximage_t *image;

  image= calloc(1,sizeof(*image));
  image->width= width;
  image->height= height;
  image->data= malloc(sizeof(uint32_t)*width*height);

  image->x11image= XCreateImage(x_display, x_visual, x_depth,
                                ZPixmap, 0, (char*) image->data,
                                width, height, 32, width*sizeof(uint32_t));
      // FIXME: error checking??
  image->bpl= image->x11image->bytes_per_line;
  image->shared= 0;
  return image;
}

/****************************************************************************
 ***************************************************************************/

static void destroy_image(ximage_t *image)
{
  XSync(x_display, False);
  if (image->shared) 
    XShmDetach(x_display, &image->shminfo);
  XDestroyImage(image->x11image);
  if (image->shared)
    shmdt(image->shminfo.shmaddr);
  free(image);
}


/************************************************************
  Putting an image on the screen is also handled differently
  for shared and non-shared images.
 ***********************************************************/

static void put_image(Drawable draw, int dest_x, int dest_y,
               ximage_t *image, int src_x, int src_y,
               int width, int height)
{
  if (src_x >= image->width ) return ;
  if (src_y >= image->height ) return ;
  if (src_x<0) { width+= src_x; src_x= 0; }
  if (src_y<0) { height+= src_y; src_y= 0; }
  
  if (width + src_x> image->width)  width= image->width-src_x;
  if (height+ src_y> image->height) height= image->height-src_y;

  if (image->shared)
    XShmPutImage(x_display, draw, DefaultGC(x_display,0),
      image->x11image, src_x, src_y, dest_x, dest_y,
      width, height, False);
  else
    XPutImage(x_display, draw, DefaultGC(x_display,0),
      image->x11image, src_x, src_y, dest_x, dest_y,
      width, height);
}

/*******************************************************
  Images in X11 can be in shared memory if the server 
  supports the "MIT-SHM" extension.
 ******************************************************/
static int can_use_shared_images()
{
  static int tried;
  static int can;
  if (!tried)
  {
    int ignore, major, minor;
    Bool pixmaps;
    tried= 1;
    can= XQueryExtension(x_display, "MIT-SHM", &ignore, &ignore, &ignore)
         && XShmQueryVersion(x_display, &major, &minor, &pixmaps);
  }
  return can;
}


/***********************************************************************
 *                                                                     *
 *                               EVENTS                                *
 *                                                                     *
 ***********************************************************************
  Exposures should be handled in a separate thread which frees up the
  main control flow to do whatever it pleases until it decides to act
  upon some events.

  I can do the following: handle events retrieval and painting by one
  thread and pass unhandled events to the main thread. When we're dealing
  with the display, disable this events thread and do your X11 calls.
  After you're done, enable it again.

  This is quite a bit of code which I don't really need. There are other
  options:

  - enabling the backing store: this isn't guaranteed to work. As I
    understand it, it's meant as an optimization which can be dropped
    any time the X server wishes to.
    (http://tronche.com/gui/x/xlib/window/attributes/backing-store.html)

  - using the background pixmap: this isn't guaranteed either. The
    X server is allowed to make a copy of the pixmap and updates to
    the original pixmap isn't necessarily propagated.
    (http://tronche.com/gui/x/xlib/window/attributes/background.html)

  - using a different connection object: this can be done. Either in
    a separate thread or a separate process. I'm leaning towards the
    process option since I don't know how well the Xlib works with
    multiple threads accessing multiple connections. Here is an excerpt
    from 
                   http://www.remlab.net/op/xlib.shtml

                  Multiple displays

       But first, a well behaved X11 error handler needs to check which
       Display generated the error, because the Xlib error handler is
       set per process, not per display. In the good old days, most
       applications could assume that they had only one display in
       their process space. The main user interface toolkit, e.g. Qt,
       GTK, would create the only display connection. But this is not
       so true anymore:

       Media frameworks such as LibVLC or gstreamer need to create their
       own displays for rendering, Rendering libraries like Cairo, EGL,
       SDL also create their own connection, Even non-graphical libraries
       like PulseAudio or D-Bus create short-lived display connections
       to identify the session that they run in, and connect to the
       right instance of the corresponding daemon.

       And then, the error handler need to pass the error down to the
       previous error handler that might have been registered by another
       component, as returned by XSetErrorHandler() earlier.

       If you are smart and fast, you might already have found a major
       flaw. To unregister an error handler, XSetErrorHandler() is called
       a second time with the previous return value as parameter, just
       like with the POSIX signal() function. And just like POSIX signal
       handlers, this cannot work unless the handlers are unregistered
       in the exact reverse order that they were registered.

                  Multiple threads

       The error handling functions are atomic, or at least can be
       configured to behave so. Even then, the error events are not
       locked. Therefore, there is always a risk that an error in one
       thread triggers an undefined situation in the error handler of
       another thread. If you think this might still work, I humbly advise
       that you refresh your knowledge on the topic of the POSIX threaded
       memory model, until you understand why it cannot operate safely.

       As the VLC bug master, I can tell that has been one of the worse
       crash source on VLC for GNU/Linux. Fortunately, we have now fixed
       the problem: We got rid of Xlib and used XCB wherever we needed
       X11 error handling. And if multiple threads are involved, there
       is more-or-less no safe way to an error handler at all.

                  Consequences

       If XInitThreads() is not called anywhere in a multi-tasked program,
       the program will usually work. But weird crashes and other race
       conditions are bound to happen over time, as Xlib maintains some
       global data that is shared across all threads and all "displays".

       If XInitThreads() is called too late, then the process will crash,
       but only some time later in obscure ways. In my experience, the
       most common case is for the process to hit a segmentation fault
       within XCloseDisplay(), which usually happens at exit.
   
   So, I should better do the process thing. The main process will create
   a pixmap and then send its ID. The exposure process will select the
   ExposureMask on the window and do its stuff. 

   It would make more sense if I had used shared pixmaps instead of 
   shared images. Right now, there are two copies: from image to pixmap
   and then pixmap to window. Later on maybe I can do that, but I don't
   need it right now.
 ************************************************************************/


/***********************************************************************
   The following forks a process and starts waiting for a window and a
   pixmap from the main process.
 ***********************************************************************/


static void child_loop();

static int fork_exposure_process()
{
  int fd[2]; 
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) 
  {
    fprintf(stderr,"socketpair failed: %s\n",strerror(errno));
    return 1;
  }
  signal(SIGCHLD, SIG_IGN);
  expopid= fork();
  if (expopid==-1)
  {
    fprintf(stderr,"fork failed: %s\n",strerror(errno));
    close(fd[0]);
    close(fd[1]);
    return 1;
  }
  if (!expopid)
  {
    close(fd[1]);
    exposock= fd[0];
    child_loop();
  }
  close(fd[0]);
  exposock=fd[1];
  return 0;
}


/***********************************************************************
  We communicate exactly once during the whole thing. We just get the
  window ID and the pixmap ID and then go to work. If the window was
  resizable, we'd create a new pixmap and send that to the child process
  as well, but this is not necessary right now. The data is encoded as:
 ***********************************************************************/
typedef struct
{
  Window window;
  Pixmap pixmap;
  int width,height;
} gfx_child_params_t;

static int send_child_params(Window window, Pixmap pixmap,
                             int width, int height)
{
  gfx_child_params_t cp;
  char *buf;
  int len,R;

  if (exposock==-1) return 1;
  cp.window= window;
  cp.pixmap= pixmap;
  cp.width= width;
  cp.height= height;

  len= sizeof(cp);
  buf= (void*) &cp;
  while(len)
  {
    R= send(exposock, buf, len,MSG_NOSIGNAL);
    if (R<0)
    {
      if (errno==EINTR) continue;
       // When we kill the child, we don't wait for it. 
       // This lets the user see that there is something
       // wrong with the update process.
      kill(expopid, SIGKILL);
      close(exposock);
      exposock= -1;
      return 1;
    }
    len-= R;
    buf+= R;
  }
  close(exposock);
  exposock= -1;
  return 0;
}

static void recv_child_params(Window *window, Pixmap *pixmap, int *w, int *h)
{
  gfx_child_params_t cp;
  char *buf;
  int len,R;

  buf= (void*) &cp;
  len= sizeof(cp);

  while(len)
  {
    R= recv(exposock,buf,len, 0);
    if (R<0)
    {
      if (errno==EINTR) continue;
      exit(1);
    }
    buf+= R;
    len-= R;
  }
  *window= cp.window;
  *pixmap= cp.pixmap;
  *w= cp.width;
  *h= cp.height;
  close(exposock);
}


/***********************************************************************
  Now we can handle the child loop.
 ***********************************************************************/
static void child_loop()
{
  Display      *display;
  XEvent        event;
  XExposeEvent *expose;
  Window        window;
  int           width,height;
  Pixmap        pixmap;

  display= XOpenDisplay(NULL);
  if (!display) exit(1);
  recv_child_params(&window, &pixmap, &width, &height);
  XSelectInput(display, window, ExposureMask | StructureNotifyMask);
  expose= &event.xexpose;
  while(1)
  {
    XNextEvent(display, &event);
    if (event.type==DestroyNotify && event.xdestroywindow.window==window)
          exit(0);
    if (event.type!=Expose) continue;
                                      // FIXME:
                                      // This zero should be something else
    XCopyArea(display, pixmap, window, DefaultGC(display,0), 
               expose->x, expose->y, expose->width, expose->height,
               expose->x, expose->y);
  }
}

static void tv2msecs(struct timeval *tv, unsigned int* msecs)
{
  if (!msecs) return ;
  *msecs= tv->tv_sec*1000 + tv->tv_usec/1000;
}

/* The event routine returns an event from the queue. It waits a maximum
   of 'msecs' milliseconds for the event. If msecs is zero, it waits 
   until an event is received. 

   The return value is one if an event is returned. It's zero if the
   return was due to timeout.

   In case of an event, *msecs is updated to reflect the remaining
   wait time.  */
   
int gfx_event(unsigned int *msecs,gfx_event_t *Revent)
{
  int x,R;
  XEvent event;
  struct timeval tv;
  fd_set rfd, xfd, wfd;

again:
  if (!msecs || !*msecs || XPending(x_display)) goto have_event;

retry:

  if (XPending(x_display)) goto have_event;

  tv.tv_sec= *msecs/1000;
  tv.tv_usec= (*msecs%1000)*1000+1;
  if (tv.tv_sec==0 && tv.tv_usec==0) return 0;

  x= ConnectionNumber(x_display);
  XSync(x_display,False);

  FD_ZERO(&rfd); FD_ZERO(&xfd); FD_ZERO(&wfd);
  FD_SET(x, &rfd); FD_SET(x, &xfd);

  R= select( x+1, &rfd, &wfd, &xfd, &tv);
  if (R<0)
  {
    if (errno==EINTR)
    {
       usleep(1000);
       tv2msecs(&tv, msecs);
       goto retry;
    }
    return -1; // FIXME: log this stuff.
  }
  if (FD_ISSET(x, &xfd)) return -1; // FIXME: log this stuff.

  tv2msecs(&tv, msecs);
  if (!FD_ISSET(x, &rfd)) goto retry;

have_event:
  XNextEvent(x_display, &event);

  switch(button_event(&event,Revent))
  {
  case 0:  break;
  case 1:  return 1;
  case -1: goto again;
  }
  
  switch(event.type)
  {
   case MotionNotify:
          if (gfx_recsel_event(&event)) goto again;
          break;
   case KeyPress:
   case KeyRelease:
          Revent->x= event.xkey.x; 
          Revent->y= event.xkey.y; 
          Revent->modifiers= event.xkey.state; 
          Revent->key= event.xkey.keycode;
          break;
   case ButtonPress:
   case ButtonRelease:
          Revent->x= event.xbutton.x; 
          Revent->y= event.xbutton.y; 
          Revent->modifiers= event.xbutton.state; 
          Revent->button= event.xbutton.button;
          break;
  }

  switch(event.type)
  {
   case KeyPress:      Revent->type= GFXEV_KEY_PRESS; break;
   case KeyRelease:    Revent->type= GFXEV_KEY_RELEASE; break;
   case ButtonPress:   Revent->type= GFXEV_BUTTON_PRESS; break;
   case ButtonRelease: Revent->type= GFXEV_BUTTON_RELEASE; break;
   case MotionNotify:  Revent->type= GFXEV_POINTER_MOVE; 
                       Revent->x= event.xmotion.x; 
                       Revent->y= event.xmotion.y; 
                       Revent->modifiers= event.xmotion.state; 
                       break;
   case ClientMessage: 
        if (event.xclient.message_type!=wm_protocols) goto again;
        if (event.xclient.data.l[0]!=wm_delete_window) goto again;
        Revent->type= GFXEV_QUIT; break;
   default:
        printf("unknown event type: %d\n", event.type);
        goto again;
  }
  return 1; 
}

void gfx_update_all()
{
  gfx_update(0,0,x_window_width, x_window_height);
}

static void clip_for_rect(int *start, int *amount, int max)
{
  int end;
  if (*amount<=0) return ;
  end= *start + *amount;
  if (end>max) end= max;
  if (*start<0) *start= 0;
  *amount= end - *start;
}

void gfx_update(int x,int y,int width,int height)
{
  clip_for_rect(&x, &width, x_window_width);
  clip_for_rect(&y, &height, x_window_height);
  if (width<=0 || height<=0) return ;
  put_image(x_pixmap, x, y, x_image, x,y, width, height); 
  XClearArea(x_display, x_window, x,y, width, height, True);
}

void gfx_put(int flags,
             unsigned int *buffer, int src_w, int src_h,
             int src_x, int src_y,
             int dst_x, int dst_y,
             int width, int height)
{
   int x,y;
   char *data;
   int bpl;

   clip_for_rect( &src_x, &width, src_w);
   clip_for_rect( &src_y, &height, src_h);
   clip_for_rect( &dst_x, &width, x_window_width);
   clip_for_rect( &dst_y, &height, x_window_height);
   if (width<=0 || height<=0) return ;

   data= (void*) x_image->data;
   bpl= x_image->bpl;
   
   if (!(flags&GFX_ALPHA)) 
   {
     for(y=0;y<height;y++)
       for(x=0;x<width;x++)
       {
         uint32_t sR, sG, sB, sP;
         uint32_t dP;
         sP= buffer[ (x+src_x) + (y+src_y)*src_w ];
         sR= (sP >> 16)&0xff;
         sG= (sP >> 8 )&0xff;
         sB= (sP      )&0xff;

         dP= (sR << x_red_shift)&x_red_mask;
         dP|= (sG << x_green_shift)&x_green_mask;
         dP|= (sB << x_blue_shift)&x_blue_mask;

         *(uint32_t*) (data+(y+dst_y)*bpl + 4*(x+dst_x))= dP;
       }
   }
   else
   {
     for(y=0;y<height;y++)
       for(x=0;x<width;x++)
     {
         uint32_t sA, sR, sG, sB, sP;
         uint32_t     dR, dG, dB, dP;

         sP= buffer[ (x+src_x) + (y+src_y)*src_w ];
         sA= (sP >> 24)&0xff;
         if (!sA) continue;
         sR= (sP >> 16)&0xff;
         sG= (sP >> 8 )&0xff;
         sB= (sP      )&0xff;

         if (sA==0xff)
         {
            dR= sR;
            dG= sG;
            dB= sB;
         }
         else
         {
            double R,V;

            dP= *(uint32_t*) (data+(y+dst_y)*bpl + 4*(x+dst_x));

            dR= (dP & x_red_mask)>>x_red_shift;
            dG= (dP & x_green_mask)>>x_green_shift;
            dB= (dP & x_blue_mask)>>x_blue_shift;

            V= sA/255.0; R= 1.0 - V;

            dR=  dR*R + sR*V;
            dG=  dG*R + sG*V;
            dB=  dB*R + sB*V;
         }

         dP= (dR << x_red_shift)&x_red_mask;
         dP|= (dG << x_green_shift)&x_green_mask;
         dP|= (dB << x_blue_shift)&x_blue_mask;

         *(uint32_t*) (data+(y+dst_y)*bpl + 4*(x+dst_x))= dP;
     }
   }

   if (flags & GFX_UPDATE)
      gfx_update(dst_x,dst_y,width,height);
}

static void checker_pattern(ximage_t *img,int N)
{
  int x,y;
  uint32_t all_mask;
  if (N<=0) N= 4;
  all_mask= x_red_mask | x_green_mask | x_blue_mask;
  for(y=0;y<img->height;y++)
   for(x=0;x<img->width;x++)
     *(uint32_t*)  ((unsigned char*) img->data + y*img->bpl + 4*x)= 
            (x%(2*N)/N ^ y%(2*N)/N)  ? all_mask : 0;
}

void gfx_close()
{
  if (x_image) destroy_image(x_image);
  XDestroyWindow(x_display, x_window);
  XFreePixmap(x_display, x_pixmap);
  if (exposock!=-1) close(exposock);
  x_display= NULL;
  x_window= None;
  x_pixmap= None;
  x_image= NULL;
  exposock= -1;
}

void gfx_get(unsigned int *buffer, int dst_w, int dst_h,
             int dst_x, int dst_y,
             int src_x, int src_y,
             int width, int height)
{
   int x,y;
   char *data;
   int bpl;

   clip_for_rect( &src_x, &width, x_window_width);
   clip_for_rect( &src_y, &height, x_window_height);
   clip_for_rect( &dst_x, &width, dst_w);
   clip_for_rect( &dst_y, &height, dst_h);
   if (width<=0 || height<=0) return ;

   data= (void*) x_image->data;
   bpl= x_image->bpl;
   
   for(y=0;y<height;y++)
     for(x=0;x<width;x++)
     {
         uint32_t sP;
         uint32_t dR, dG, dB, dP;

         sP= *(uint32_t*) (data+(y+src_y)*bpl + 4*(x+src_x));

         dR= (sP & x_red_mask)>>x_red_shift;
         dG= (sP & x_green_mask)>>x_green_shift;
         dB= (sP & x_blue_mask)>>x_blue_shift;
 
         dP= (0xff<<24) | (dR<<16) | (dG<<8) | (dB);
         buffer[ (x+dst_x) + (y+dst_y)*dst_w ]= dP;
     }
}

/***********************************************************************
 *                                                                     *
 *                       Rectangular Selections                        *
 *                                                                     *
 ***********************************************************************
  There are some things to do:
    - only update the borders of the rectangle, not the whole thing.
    - constrain the selection to a given area.
 ***********************************************************************/

typedef struct { int x,y; } gfx_point_t;
static int recsel;
static gfx_point_t recselpt[4];
static void draw_recsel();

static int imin2(int a,int b) { return a < b ? a : b; }
static int imax2(int a,int b) { return a > b ? a : b; }

static int imin3(int a,int b,int c) { return imin2(imin2(a,b),c); }
static int imax3(int a,int b,int c) { return imax2(imax2(a,b),c); }

void gfx_recsel_begin(int x,int y,int cox,int coy,int cow,int coh)
{
  recselpt[0].x= x;
  recselpt[0].y= y;
  recselpt[1]= recselpt[0];
  recsel= 1;

  if (x<0 || x>=x_window_width) x= 0;
  if (y<0 || y>=x_window_height) y= 0;
  if (cow<0) cow= x_window_width - cox;
  if (coh<0) coh= x_window_height - coy;
  recselpt[2].x= cox;       recselpt[2].y= coy;
  recselpt[3].x= cox+cow-1; recselpt[3].y= coy+coh-1;
}

void gfx_recsel_end(int flags,int *Rx1, int *Ry1, int *Rx2, int *Ry2)
{
  int minx,miny,maxx,maxy,Z;

  draw_recsel();
  minx= imin2( recselpt[0].x, recselpt[1].x);
  maxx= imax2( recselpt[0].x, recselpt[1].x);
  miny= imin2( recselpt[0].y, recselpt[1].y);
  maxy= imax2( recselpt[0].y, recselpt[1].y);
  gfx_update(minx, miny, maxx-minx+1, maxy-miny+1);

  Rx1= Rx1 ? Rx1 : &Z;
  Ry1= Ry1 ? Ry1 : &Z;
  Rx2= Rx2 ? Rx2 : &Z;
  Ry2= Ry2 ? Ry2 : &Z;

  if (flags & GFX_ORDER)
  {
    *Rx1= minx; *Rx2= maxx;
    *Ry1= miny; *Ry2= maxy;
  }
  else
  {
    *Rx1= recselpt[0].x;
    *Ry1= recselpt[0].y;
    *Rx2= recselpt[1].x;
    *Ry2= recselpt[1].y;
  }
  recsel= 0;
}

static int gfx_recsel_event(XEvent *event)
{
  gfx_point_t P;
  int x,y,minx,miny,maxx,maxy;
  if (!recsel || event->type!=MotionNotify) return 0;
  x= event->xmotion.x;
  y= event->xmotion.y;

  x= imin2(imax2(x,recselpt[2].x), recselpt[3].x); 
  y= imin2(imax2(y,recselpt[2].y), recselpt[3].y); 

  if (x==recselpt[1].x && y==recselpt[1].y) return 1;
  
  draw_recsel();
  minx= imin3( recselpt[0].x, recselpt[1].x, x);
  maxx= imax3( recselpt[0].x, recselpt[1].x, x);

  miny= imin3( recselpt[0].y, recselpt[1].y, y);
  maxy= imax3( recselpt[0].y, recselpt[1].y, y);

  recselpt[1].x= x;
  recselpt[1].y= y;
  draw_recsel();
  gfx_update(minx, miny, maxx-minx+1, maxy-miny+1);
  return 1;
}

static void gfx_invert_hline(int x1,int x2, int y)
{
  uint32_t all_mask;
  int bpl;
  char *data;

  if (y<0 || y>=x_window_height || x1>x2 || x1>=x_window_width) return ;
  if (x2>=x_window_width) x2= x_window_width-1;

  all_mask= x_red_mask | x_green_mask | x_blue_mask;
  bpl= x_image->bpl;
  data= (void*) x_image->data;
  for(;x1<=x2;x1++)
    *(uint32_t*) (data + y*bpl + x1*4) ^= all_mask;
}

static void gfx_invert_vline(int x,int y1, int y2)
{
  uint32_t all_mask;
  int bpl;
  char *data;

  if (x<0 || x>=x_window_width || y1>y2 || y1>=x_window_height) return ;
  if (y2>=x_window_height) y2= x_window_height-1;

  all_mask= x_red_mask | x_green_mask | x_blue_mask;
  bpl= x_image->bpl;
  data= (void*) x_image->data;
  for(;y1<=y2;y1++)
    *(uint32_t*) (data + y1*bpl + x*4) ^= all_mask;
}

static void draw_recsel()
{
  int x,y,w,h,C;

  x= imin2(recselpt[0].x, recselpt[1].x);
  y= imin2(recselpt[0].y, recselpt[1].y);

  w= imax2(recselpt[0].x, recselpt[1].x) - x+1;
  h= imax2(recselpt[0].y, recselpt[1].y) - y+1;

  C= (w<2) ? 0 : 1;
  C|= (h<2) ? 0 : 2;
  switch(C)
  {
  case 0: break ;
  case 1: gfx_invert_hline(x, x+w-1, y); break ;
  case 2: gfx_invert_vline(x, y, y+h-1); break ;
  case 3: gfx_invert_hline(x, x+w-1, y);
          gfx_invert_hline(x, x+w-1, y+h-1);
          gfx_invert_vline(x, y+1, y+h-2);
          gfx_invert_vline(x+w-1, y+1, y+h-2);
          break;
  }
}


/***********************************************************************
  More things to do:
  - text, maybe with some embedded font only. I have bfont for that.
  - buttons.
  - proper handling of the keyboard.
  - mouse position indicator in pixels
 ***********************************************************************/

#include "font1.cx"
#include "font2.cx"
#include "font3.cx"

static void get_text_data(int font, char *str, 
                          int *Rw, int *Rh,int *Ra,unsigned char **Rd)
{
  uint32_t *wtr;
  size_t L= strlen(str);
  wtr= alloca((L+1)*sizeof(uint32_t));
  utf8_decode_string((uint8_t*) str, L, wtr, &L);

  if (font<1) font=1;
  if (font>2) font=2;
  switch(font)
  {
  case 1: font1_render(wtr, L, Rw, Rh, Ra, Rd); break;
  case 2: font2_render(wtr, L, Rw, Rh, Ra, Rd); break;
  }
}

void gfx_text_size(int font, char *str, int *W, int *H)
{
  get_text_data(font, str, W, H, NULL, NULL);
}

static uint32_t blend_usr_usr(uint32_t alpha, uint32_t fg, uint32_t bg)
{
  uint32_t fR, fG, fB; 
  uint32_t bR, bG, bB; 
  uint32_t pR, pG, pB; 
  double A,B;

  fR= (fg>>16)&0xff; fG= (fg>>8)&0xff; fB= fg&0xff;

  if (alpha==0xff) { pR= fR; pG= fG; pB= fB; goto OK;}

  bR= (bg>>16)&0xff; bG= (bg>>8)&0xff; bB= bg&0xff;
  if (alpha==0) { pR= bR; pG= bG; pB= bB; goto OK; }

  A= (double) alpha / 255.0;
  B= 1.0-A;

  pR= A*fR + B*bR;
  pG= A*fG + B*bG;
  pB= A*fB + B*bB;

OK:
  return   ((pR<<x_red_shift)&x_red_mask)
         | ((pG<<x_green_shift)&x_green_mask)
         | ((pB<<x_blue_shift)&x_blue_mask);
}

static uint32_t blend_usr_svr(uint32_t alpha, uint32_t fg, uint32_t bg)
{
  uint32_t fR, fG, fB; 
  uint32_t bR, bG, bB; 
  uint32_t pR, pG, pB; 
  double A,B;

  if (alpha==0) return bg;

  fR= (fg>>16)&0xff; fG= (fg>>8)&0xff; fB= fg&0xff;

  if (alpha==0xff) { pR= fR; pG= fG; pB= fB; goto OK;}

  bR= (bg&x_red_mask)>>x_red_shift; 
  bG= (bg&x_green_mask)>>x_green_shift;
  bB= (bg&x_blue_mask)>>x_blue_shift;

  A= (double) alpha / 255.0;
  B= 1.0-A;

  pR= A*fR + B*bR;
  pG= A*fG + B*bG;
  pB= A*fB + B*bB;

OK:
  return   ((pR<<x_red_shift)&x_red_mask)
         | ((pG<<x_green_shift)&x_green_mask)
         | ((pB<<x_blue_shift)&x_blue_mask);
}

static void blend_graymap(int flags, int fg, int bg,
                          unsigned char *src, int src_w, int src_h,
                          int src_x, int src_y,
                          int dst_x, int dst_y, int width, int height)
{
  int x,y;
  uint32_t A;

  clip_for_rect( &src_x, &width, src_w);
  clip_for_rect( &src_y, &height, src_h);
  clip_for_rect( &dst_x, &width, x_window_width);
  clip_for_rect( &dst_y, &height, x_window_height);
  if (width<=0 || height<=0) return ;

  if (flags&GFX_BACKGND) // blend with fg,bg pair
  {
    for(y=0;y<height;y++)
      for(x=0;x<width;x++)   
      {
        A= src[(x+src_x) + (y+src_y)*src_w];
        x_image_data[y+dst_y][x+dst_x]= blend_usr_usr(A,fg, bg);
      }
  } 
  else
  {
    for(y=0;y<height;y++)
      for(x=0;x<width;x++)   
      {
        A= src[(x+src_x) + (y+src_y)*src_w];
        x_image_data[y+dst_y][x+dst_x]= 
             blend_usr_svr(A,fg, x_image_data[y+dst_y][x+dst_x]);
      }
  } 

  if (flags&GFX_UPDATE)
    gfx_update(dst_x, dst_y, width, height);
}

void gfx_print(int flags, int font, int fg, int bg, int x, int y, 
               char *str,int *Rw, int *Rh)
{
  int w,h; unsigned char *data;
  get_text_data(font, str, &w, &h, NULL, &data);

  blend_graymap(flags, fg, bg, data, w,h, 0,0, x,y,w,h);
  if (data) free(data);
  if (Rw) *Rw= w;
  if (Rh) *Rh= h;
}


/***********************************************************************
 *                                                                     *
 *                            B U T T O N S                            *
 *                                                                     *
 ***********************************************************************
 *  Each button is represented by the following struct:                *
 ***********************************************************************/

typedef struct
{
  int x,y,w,h;
  uint32_t *pic[3];
  int state;  //0 = normal, 1= depressed 2= disabled
  int action;
} gui_button_t;

static int n_buttons, s_buttons;
static gui_button_t *buttons;

static uint32_t button_border_s= 0xaaaaa3, button_border,
                button_light_s=  0xffffff, button_light,
                button_dark_s=   0xd5d5d1, button_dark,
                button_grad1_s=  0xfcfcfb, button_grad1,
                button_grad2_s=  0xdbdbd7, button_grad2,
                button_grad3_s=  0xcccac9, button_grad3,
                button_grad4_s=  0xe9e8e7, button_grad4,
                button_fg_s=     0,        button_fg; 

static uint32_t color_s(uint32_t uR, uint32_t uG, uint32_t uB)
{
  return ((uR<<x_red_shift)&x_red_mask)
        |((uG<<x_green_shift)&x_green_mask)
        |((uB<<x_blue_shift)&x_blue_mask);
}

static uint32_t u_red(uint32_t C) { return (C>>16)&0xff; }
static uint32_t u_green(uint32_t C) { return (C>>8)&0xff; }
static uint32_t u_blue(uint32_t C) { return (C)&0xff; }

static uint32_t color_u_to_s(uint32_t U)
{
  return color_s(u_red(U), u_green(U), u_blue(U));
}

static unsigned int blend_chan(unsigned int F, double A, unsigned int B)
{
  return  (unsigned int) ((F&0xff)*A + (B&0xff)*(1-A));
}


static unsigned int blend(uint32_t F, double A, uint32_t B)
{
  return color_s(
            blend_chan( u_red(F), A, u_red(B)),
            blend_chan( u_green(F), A, u_green(B)),
            blend_chan( u_blue(F), A, u_blue(B))
                );
}

static void grad_fill(int x,int y, int w, int h, uint32_t start, uint32_t end)
{
  int fx, fy, fw, fh,x0,y0,dx,dy;
  fx= x; fy= y; fw= w; fh= h;

  clip_for_rect(&x, &w, x_window_width);
  clip_for_rect(&y, &h, x_window_height);
  if (w<=0 || h<=0) return;

  for(y0=0;y0<h;y0++)
    for(x0=0;x0<w;x0++)
    {
      dx= (x0+x)-fx; dy= (y0+y)-fy;
      x_image_data[y0+y][x0+x]= blend(start, 1.0-(double)(dx+dy)/(fw+fh), end);
    }
}


static void init_button_colors()
{
  button_border= color_u_to_s(button_border_s);
  button_light= color_u_to_s(button_light_s);
  button_dark= color_u_to_s(button_dark_s);
  button_grad1= color_u_to_s(button_grad1_s);
  button_grad2= color_u_to_s(button_grad2_s);
  button_grad3= color_u_to_s(button_grad3_s);
  button_grad4= color_u_to_s(button_grad4_s);
  button_fg= color_u_to_s(button_fg_s);
}

static void hline(int x1,int x2, int y,uint32_t color)
{
  if (y<0 || y>=x_window_height || x1>x2 || x1>=x_window_width) return ;
  if (x2>=x_window_width) x2= x_window_width-1;

  for(;x1<=x2;x1++)
    x_image_data[y][x1]= color;
}

static void vline(int x,int y1, int y2,uint32_t color)
{
  if (x<0 || x>=x_window_width || y1>y2 || y1>=x_window_height) return ;
  if (y2>=x_window_height) y2= x_window_height-1;

  for(;y1<=y2;y1++)
    x_image_data[y1][x]= color;
}

static void button_render(int x,int y,int w,int h,
                              int tw, int th, unsigned char *data,
                              uint32_t grad1, uint32_t grad2,
                              uint32_t light, uint32_t dark)
{
  int copy_w, copy_h,src_x,src_y, text_x, text_y;
  int aw, ah;

  aw= w-4; ah= h-4;
  grad_fill(x+2,y+2,aw,ah, grad1, grad2);

  if (tw>aw) { src_x= (tw-aw)/2; copy_w= aw; text_x= x+2; }
        else { src_x= 0; copy_w= tw; text_x= x+2+(aw-tw)/2; }
  if (th>ah) { src_y= (th-ah)/2; copy_h= ah; text_y= y+2; }
        else { src_y= 0; copy_h= th; text_y= y+2+(ah-th)/2; }

  blend_graymap(0, button_fg, 0, data, tw,th, src_x,src_y,
                                 text_x,text_y,copy_w,copy_h);

  hline(x, x+w-1, y, button_border);
  hline(x, x+w-1, y+h-1, button_border);
  vline(x, y, y+h-1, button_border);
  vline(x+w-1, y, y+h-1, button_border);

  hline(x+1, x+w-3, y+1, light); 
  vline(x+1, y+1, y+h-3, light);

  hline(x+1, x+w-2, y+h-2, dark);
  vline(x+w-2, y+1, y+h-2, dark);
}

static uint32_t *get_server_pix(int src_x,int src_y,int w,int h)
{
  uint32_t *D; int dw;
  int x,y;
  D= calloc(1,sizeof(uint32_t*)*w*h);
  dw= w;
  clip_for_rect(&src_x, &w, x_window_width);
  clip_for_rect(&src_y, &h, x_window_height);

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      D[x+y*dw]= x_image_data[y+src_y][x+src_x];
  return D;
}

static void put_server_pix(uint32_t *D,int dst_x,int dst_y,int w, int h)
{
  int dw;
  int x,y;
  dw= w;
  clip_for_rect(&dst_x, &w, x_window_width);
  clip_for_rect(&dst_y, &h, x_window_height);

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      x_image_data[y+dst_y][x+dst_x]= D[x+y*dw];
}

static void *grow_array(int N, int *S, void *ptr, size_t eltsiz)
{
  if (N<*S) return ptr;
  if (!*S)
  {
    *S= 20;
    ptr= malloc(*S*eltsiz);
  }
  else
  {
    *S+= *S/2;
    ptr= realloc(ptr, *S*eltsiz);
  }
  return ptr;
}


int gfx_button(int flags,int x,int y, int w, int h, char *str, int action)
{
  int tw,th; unsigned char *data;
  int bno;

  buttons= grow_array(n_buttons, &s_buttons, buttons, sizeof(buttons[0]));
  bno= n_buttons++;
  buttons[bno].action= action;
  buttons[bno].state= 0;
  buttons[bno].x= x;
  buttons[bno].y= y;
  buttons[bno].w= w;
  buttons[bno].h= h;

  buttons[bno].pic[2]= get_server_pix(x,y,w,h);

  get_text_data(1, str, &tw, &th, NULL, &data);

  button_render(x,y,w,h, tw, th, data, 
                button_grad3, button_grad4, 
                button_dark, button_light);

  buttons[bno].pic[1]= get_server_pix(x,y,w,h);
  
  button_render(x,y,w,h, tw, th, data, 
                button_grad1, button_grad2, 
                button_light, button_dark);

  buttons[bno].pic[0]= get_server_pix(x,y,w,h);

  free(data);

  if (flags&GFX_UPDATE)
    gfx_update(x,y,w,h);

  return bno;
}

static int depressed_button= -1;

static void button_set_state(int bno, int state)
{
  if (bno>=n_buttons || bno<0) return ;
  if (state<0 || state>2) state= 0;
  buttons[bno].state= state;

  put_server_pix( buttons[bno].pic[buttons[bno].state],
                  buttons[bno].x, buttons[bno].y,
                  buttons[bno].w, buttons[bno].h);

  gfx_update(buttons[bno].x, buttons[bno].y, buttons[bno].w, buttons[bno].h);
}

static int range_test(int v, int s, int w) { return v>=s && v<(s+w); }

static int button_event(XEvent *event, gfx_event_t *R)
{
  int bno;
  if (event->type==ButtonPress)
  {
     if (event->xbutton.button!=1) return 0;
     if (depressed_button!=-1)
     {
       // we somehow missed the release event
       button_set_state(depressed_button,0);
       depressed_button= -1;
     }
     for(bno=0;bno<n_buttons;bno++)
       if (range_test(event->xbutton.x,buttons[bno].x,buttons[bno].w) &&
           range_test(event->xbutton.y,buttons[bno].y,buttons[bno].h))
             break;
     if (bno==n_buttons) return 0;
     if (buttons[bno].state==2) return 0;
     button_set_state(bno,1);
     depressed_button= bno;
     return -1;
  }
  if (event->type==ButtonRelease)
  {
     if (event->xbutton.button!=1) return 0;
     if (depressed_button==-1)
       return 0;
     bno= depressed_button;
     depressed_button= -1;
     button_set_state(bno,0);
     if (bno<n_buttons)
     { R->type= buttons[bno].action; return 1; }
     return -1;
  }
  return 0;
}

void gfx_button_enable(int btn,int a)
{
  if (btn<0 || btn>=n_buttons) return ;
  if (a) button_set_state(btn, 0);
    else button_set_state(btn, 2);
}

void gfx_fill_rect(int flags, int dst_x,int dst_y,
                   int w,int h, unsigned int color)
{
  int x,y;
  uint32_t sv;
  clip_for_rect(&dst_x, &w, x_window_width);
  clip_for_rect(&dst_y, &h, x_window_height);
  if (w<=0 || h<=0) return ;
  sv= color_u_to_s(color);
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
     x_image_data[y+dst_y][x+dst_x]= sv;
  if (flags&GFX_UPDATE) gfx_update(dst_x, dst_y, w,h);
}


/***********************************************************************
 *                                                                     *
 *                             L A B E L S                             *
 *                                                                     *
 ***********************************************************************/

typedef struct 
{
  int font, fg, bg, x,y,w,h,flags;
} gui_label_t;

static int n_labels, s_labels;
static gui_label_t *labels;

static void draw_label(int flags, gui_label_t *G, char *str)
{
  int tw, th; unsigned char *td;
  int dst_x, dst_y, copy_w, copy_h, src_x, src_y;
  
  get_text_data(G->font, str, &tw, &th, NULL, &td);
  gfx_fill_rect(0, G->x, G->y, G->w, G->h, G->bg);

  if (th<=G->h)
  {
    src_y= 0; dst_y= G->y + (G->h-th)/2; copy_h= th;
  }
  else
  {
    src_y= (th-G->h)/2; dst_y= G->y; copy_h= G->h;
  }

  if (tw<G->w)
  {
    src_x= 0; copy_w= tw;
    if (G->flags&GFX_JUST_CENTER)
      dst_x= G->x+(G->w-tw)/2;
    else if (G->flags&GFX_JUST_RIGHT)
      dst_x= G->x+G->w-tw;
    else
      dst_x= G->x;
  }
  else
  {
    dst_x= G->x; copy_w= G->w;
    if (G->flags&GFX_CUT_LEFT)
      src_x= tw - G->h;
    else
      src_x= 0;
  }

  blend_graymap(GFX_BACKGND, G->fg, G->bg, td, tw, th,
                src_x, src_y, dst_x, dst_y, copy_w, copy_h);

  if (flags&GFX_UPDATE) gfx_update(G->x,G->y,G->w,G->h);
}

int gfx_label(int flags, int font, int fg, int bg,
              int x,int y, int w, int h, char *str)
{
  gui_label_t *G;
  int R;
  labels= grow_array(n_labels, &s_labels, labels, sizeof(labels[0]));
  G= labels + n_labels;
  R= n_labels;
  n_labels++;
  G->font= font;
  G->fg= fg;
  G->bg= bg;
  G->x= x;
  G->y= y;
  G->w= w;
  G->h= h;
  G->flags= (GFX_JUST_RIGHT | GFX_JUST_CENTER | GFX_CUT_LEFT) & flags;
  draw_label(flags,G, str);
  return R;
}

void gfx_label_set_text(int flags,int labno, char *str)
{
  if (labno<0 || labno>=n_labels) return ;
  draw_label(flags, labels + labno,str);
}
